from .application_layer import tic
from .autodetect import autodetect, detect


class Tic:
    def __init__(self, device_path, mode, sep):
        self.device_path = device_path
        self.serial_number = None
        self.groups = {}
        self._mode = mode
        self._sep = sep

    @classmethod
    async def create(cls, device_path):
        params = await detect(device_path)
        if not params:
            raise RuntimeError(f'No tic found at {device_path}')
        return cls(device_path, params['mode'], params['sep'])

    async def async_update(self):
        self.groups = await anext(tic(self.device_path, self._mode, self._sep))
        self.serial_number = self.groups.get("Adresse Secondaire du Compteur", self.groups.get("Adresse du compteur"))
        return self.groups

    @classmethod
    async def discover(cls):
        tics_params = await autodetect()
        return [cls(p['port'], p['mode'], p['sep']) for p in tics_params]



import serial_asyncio


HISTORIQUE = 'H'
STANDARD = 'S'
BAUDRATES = {
    HISTORIQUE: 1200,
    STANDARD: 9600
}
END_FRAME = b'\x03'


async def serial_line(port, mode):
    reader, writer = await serial_asyncio.open_serial_connection(
        url=port,
        baudrate=BAUDRATES[mode],
        bytesize=serial_asyncio.serial.SEVENBITS,
        parity=serial_asyncio.serial.PARITY_EVEN,
        stopbits=serial_asyncio.serial.STOPBITS_ONE,
        rtscts=1
    )
    return reader, writer


async def frame_iterator(reader):
    await reader.readuntil(END_FRAME)  # Skip truncated frame
    while True:
        frame = await reader.readuntil(END_FRAME)
        yield frame.decode('ascii')



from unittest import TestCase
from datetime import datetime

from enedis_tic.link_layer import SumChecker, data_set, frame


class FrameTest(TestCase):
    def test_no_dataset(self):
        with self.assertRaises(RuntimeError) as cm:
            frame('')
        self.assertEqual('Empty Frame', cm.exception.args[0])

    def test_minimal(self):
        actual = frame('\nLABEL\tDATA\t#\r')
        expected = {'LABEL': 'DATA'}
        self.assertEqual(expected, actual)

    def test_many_data_set(self):
        actual = frame('\nLABEL\tDATA\t#\r\nLABEL2\tDATA2\tP\r')
        expected = {'LABEL': 'DATA', 'LABEL2': 'DATA2'}
        self.assertEqual(expected, actual)


class DataSetTest(TestCase):
    def test_valid_historique(self):
        actual = data_set('LABEL\tDATA\t#')
        expected = {'label': 'LABEL', 'data': 'DATA'}
        self.assertEqual(expected, actual)

    def test_valid_historique_older(self):
        actual = data_set('LABEL DATA :', sep=' ')
        expected = {'label': 'LABEL', 'data': 'DATA'}
        self.assertEqual(expected, actual)

    def test_valid_standard(self):
        actual = data_set('LABEL\tH081225223518\tDATA\t$')
        expected = {'label': 'LABEL', 'data': 'DATA', 'datetime': datetime(2008, 12, 25, 22, 35, 18)}
        self.assertEqual(expected, actual)

    def test_invalid_checksum(self):
        with self.assertRaises(RuntimeError) as cm:
            data_set('LABEL DATA f', sep=' ')
        self.assertEqual('Invalid checksum: "LABEL DATA f"', cm.exception.args[0])

    def test_no_data(self):
        with self.assertRaises(RuntimeError) as cm:
            data_set('')
        self.assertEqual('Invalid DataSet: ""', cm.exception.args[0])

    def test_only_label(self):
        with self.assertRaises(RuntimeError) as cm:
            data_set('aA')
        self.assertEqual('Invalid DataSet: "aA"', cm.exception.args[0])


class ChecksumTest(TestCase):
    def test_verify_sp_older(self):
        self.assertTrue(SumChecker('LABEL DATA :').verify())

    def test_verify_sp_newer(self):
        self.assertTrue(SumChecker('LABEL DATA Z').verify())

    def test_verify_ht_older(self):
        self.assertTrue(SumChecker('LABEL\tDATA\t#').verify())

    def test_verify_ht_newer(self):
        self.assertTrue(SumChecker('LABEL\tDATA\t,').verify())

    def test_verify_datetime(self):
        self.assertTrue(SumChecker('LABEL\tH081225223518\tDATA\t$').verify())

    def test_verify_fail(self):
        self.assertFalse(SumChecker('LABEL DATA f').verify())

    def test_verify_invalid_data_set(self):
        """The given data_set is invalid, but we don't care of its validity at that level."""
        self.assertTrue(SumChecker('aA').verify())



import asyncio

from . import physical_layer
from serial.tools.list_ports import comports


async def autodetect():
    tasks = []
    for port in comports():
        tasks.append(asyncio.create_task(detect(port.device)))
    await asyncio.gather(*tasks)
    return [t.result() for t in tasks if t.result()]


async def detect(port):
    for mode in ('H', 'S'):
        try:
            res = await asyncio.wait_for(_detect_port(port, mode), 3)
        except Exception:
            continue
        else:
            return res
    return None


async def _detect_port(port, mode):
    reader, writer = await physical_layer.serial_line(port, mode)
    async for raw_frame in physical_layer.frame_iterator(reader):
        return {
            'port': port,
            'mode': mode,
            'sep': '\t' if '\t' in raw_frame else ' '
        }
    writer.close()



from unittest import TestCase

from enedis_tic.application_layer import parse_frame


class AdcoTest(TestCase):
    def test_adco(self):
        expected = {
            "Adresse du compteur": "123456789ABC"
        }
        actual = parse_frame({'ADCO': '123456789ABC'})
        self.assertEqual(expected, actual)


class AdscTest(TestCase):

    def test_adsc(self):
        expected = {
            "Adresse secondaire du compteur": "846119789ABC",
            "Constructeur": "SAGEM / SAGEMCOM",
            "Appareil": "Compteur monophasé 60A LINKY - généralisation G3 - arrivée haute",
            "Année de fabrication": 2019,
            "Numéro de série": "789ABC"
        }
        actual = parse_frame({'ADSC': '846119789ABC'})
        self.assertEqual(expected, actual)

    def test_constructeur_none(self):
        actual = parse_frame({'ADSC': 'XX6119789ABC'})
        self.assertIsNone(actual['Constructeur'])

    def test_appareil_none(self):
        actual = parse_frame({'ADSC': '84XX19789ABC'})
        self.assertIsNone(actual['Appareil'])

    def test_invalid_year(self):
        actual = parse_frame({'ADSC': '8461XX789ABC'})
        self.assertIsNone(actual.get('Année de fabrication'))


class OptarifTest(TestCase):
    def test_base(self):
        self.assertEqual({'Option tarifaire choisie': 'Base'}, parse_frame({'OPTARIF': 'BASE'}))

    def test_hc(self):
        self.assertEqual({'Option tarifaire choisie': 'Heures Creuses'}, parse_frame({'OPTARIF': 'HC..'}))

    def test_tempo(self):
        self.assertEqual({'Option tarifaire choisie': 'Tempo'}, parse_frame({'OPTARIF': 'BBR.'}))

    def test_ejp(self):
        self.assertEqual({'Option tarifaire choisie': 'EJP'}, parse_frame({'OPTARIF': 'EJP.'}))


class IsouscTest(TestCase):
    def test_isousc(self):
        actual = parse_frame({'ISOUSC': '12'})
        self.assertTrue('Intensité souscrite' in actual)
        self.assertEqual(12, actual['Intensité souscrite'].value)
        self.assertEqual('A', actual['Intensité souscrite'].unit)

